#!/bin/perl
use strict;
use KD;


=pod

=head1 NAME

lof.pl - Perl implemention for the LOF(Local Outlier Factor) algorithm.

=head1 SYNOPSIS

lof.pl k InputFile

INPUT
k:	k nearest neighbors, k>=1, k is an integer. 
	If the query point is a point of the original data file, the 1st nearest point is the query point itself.
	
	K is the minimum number of objects a "cluster" has to contain. 
	K is the maximum number of "close by" objects than can potentially be local outliers.
	The LOF paper suggests k is between 10 and 20.
	
InputFile: Each line is a point. Attributes are seperated by a space ' '. 


OUTPUT
lof.txt: Each line is a point's lof value for the input file. The higher the LOF value, the stronger the outlier.	



=cut


my ($k, $input, $myfolder);
if (@ARGV!=3)
{
	print "Usage:\n";
	print "\tlof.pl k inputfile\n\n";
	print "\twhere:\n";
	print "\tk		k nearest neighbors (MinPts)\n";
	print "\tinputfile	data points file\n";

	exit;
}
else
{
	if ($ARGV[0]=~/[1-9]+/)
	{
		$k = $ARGV[0];
		$myfolder = $ARGV[2];
	}
	else
	{
		print "k nearest neighbors (MinPts) is an integer.\n";
		exit;
	}
	
	$input = $ARGV[1];
	open (INPUT, "<$input") || die ("cannot open \"$input\": $!");		
}


print "k: $k \n";
my $kdlog = 'lof/' . $input . '_scores_' . $k . '.txt';
print "$kdlog \n";
my $start_user = time();
# insert the original points into the kd tree
my $kdTree = new Tree::KD;
my @points;
my $numPoints=0;
print "inserting kd tree...\n";
while (my $line = <INPUT>)
{
	chomp ($line);
	@{$points[$numPoints]} = split(' ', $line);	
	$kdTree->insert($numPoints,\@{$points[$numPoints]});
	$numPoints++;
}
close INPUT;

# search k nearest neighbors
my %knn;
my %k_dist;
printf "searching k nearest neighbors... $numPoints \n";

for(my $i=0;$i<$numPoints;$i++)
{
	#print "remaining: ($numPoints-$i)";

	#if ( ($numPoints-$i) % 1000 == 0){
	#	print "remaining: ($numPoints-$i)";
	#}
	my $snum = 200;
	while(1)
	{
		#print "num: $snum\n";

		my $neighbors = $kdTree->nearest($points[$i], $snum);
		pop @$neighbors; # the nearest point is itself
	
		my $numtrue = 0;
		my $okflag = 0;
		for(my $j=@$neighbors;$j>0;$j--)
		{
			my $nnb = pop @$neighbors;
			my $index = $nnb->[1];
			my $dist = sqrt($nnb->[0]);
		
			#print "dist: $j $dist \n";
			if ($dist == 0){
				#print "YES?\n";
				next;
			}else{
				$knn{$i}->{$index} = $dist; 	
				#if $j==$k;	# k distance of an object $query is the biggest k nn of $query.
				$numtrue += 1;
				if ($numtrue == $k){
					$okflag = 1;
					$k_dist{$i}=$dist;
					last;
				}
			}
		}


		if ($okflag == 1){
			#print "NUUUUUMMMM: $snum\n";
			last;
		}else{
			#print "take a look: $snum\n";
			$snum = $snum + 100;
		}
	  }	
}

# reachability distance of an object p w.r.t. object o.
printf "LOF......\n";
my %reach_dist;
foreach my $point (keys %knn)
{
	foreach my $neighbor (keys %{$knn{$point}})
	{
		my $dist=$knn{$point}->{$neighbor};
		$reach_dist{$point}->{$neighbor} = max($k_dist{$neighbor},$dist);
	}
}

# local reachability density of an object p
my %lrd_p;
foreach my $point (keys %reach_dist)
{
	my $total=0;
	foreach my $neighbor (keys %{$reach_dist{$point}})
	{
		$total += $reach_dist{$point}->{$neighbor};
	}
	$lrd_p{$point}=$k/$total;
}

# local outlier factor of an object p
my %lof;
foreach my $point (keys %reach_dist)
{
	my $total=0;
	foreach my $neighbor (keys %{$reach_dist{$point}})
	{
		$total += $lrd_p{$neighbor}/$lrd_p{$point};
	}
	$lof{$point}=$total/$k;
}

my $end_user = time();
my $time = $end_user - $start_user;
printf "LOF took %.2f seconds user time.\n\n", $time;

 

#my $kdlog = $folderr . '/lof_' . $k . '.txt'; 
open (LOFLOG, ">$kdlog") || die ("cannot open \"$kdlog\": $!");

for(my $i=0;$i<$numPoints;$i++)
{
	printf LOFLOG "%.5f\n",$lof{$i};
}
close LOFLOG;

sub max 
{
	my ($x, $y) = @_;
	return $x>$y ? $x : $y; 
}


__END__




